from lookml import lkml
from lookml.lib.lang import \
    ws,valid_name,possible_view_str,\
    possible_field_str,parse_references,\
    props,OperationError,DuplicatePrimaryKey,\
    CoexistanceError
import copy, re

import warnings
from typing import NewType, Any, Generator, Union, Tuple, Type
OMIT_DEFAULTS = False
DB_FIELD_DELIMITER_START = '`' 
DB_FIELD_DELIMITER_END = '`'
# LOOKML_DASHBOARDS = False
def omit_defaults(f):
    def wrapper(*args,**kwargs):
        if OMIT_DEFAULTS and args[0]._is_default():
            return ''
        else:            
            return f(*args,**kwargs)
    return wrapper

def possible_obj_str(
    s: str, 
    # t: Union[View,Field,Explore,Model,Manifest]) -> bool:
    t: Any) -> bool:
    """
    Check if the string matches the pattern of a specific object type

    Args:
        s (str): the string to check
        t (Union[View, Field, Explore, Model, Manifest]): the object type / class

    Returns:
        bool: is it a string of that object type
    """
    if isinstance(t,View):
        return re.match(ws.view_pattern,s)
    elif isinstance(t,Field):
        return re.match(ws.field_pattern,s)
    elif isinstance(t,Explore):
        return re.match(ws.explore_pattern,s)
    elif isinstance(t,Model):
        return re.match(ws.model_pattern,s)
    elif isinstance(t,Manifest):
        return re.match(ws.manifest_pattern,s)
    else:
        return False

def snakeCase(string: str) -> str:
    str1 = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', string)
    return re.sub('([a-z0-9])([A-Z])', r'\1_\2', str1).lower()

def removeSpace(string: str) -> str:  # removing special character / [|]<>,.?}{+=~!$%^&*()-
    return re.sub(r'(\s|/|\[|\]|\||\,|<|>|\.|\?|\{|\}|#|=|~|!|\+|\$|\%|\^|\&|\*|\(|\)|\-|\:)+', r'', string)

def lookCase(string: str) -> str:
    return removeSpace(snakeCase(string))

class prop(object):
    def __init__(self, 
            key: str, value: any, 
            parent: any, conf: dict={}):
        self.key = key
        self.value = value
        self.parent = parent
        self.conf = conf
    @omit_defaults
    def __str__(self):
        indent = ws.nl + (self.conf['indent'] * ws.s)
        if self.parent._dense():
            return f' {self.key}: {self.value}'
        else:
            return f'{indent}{self.key}: {self.value}'

    def __getattr__(self,key):
        if key in self.__dict__.keys():
            self.__dict__.__getitem__(key)
        else:
            return None

    def __contains__(self,item): return item in self.value
    def _type(self): return self.key
    def _dense(self): return len(self.value) <= ws.dense_str_len
    def _is_default(self): return True if self.conf['default_value'] == self.value else False

class lookml(object):
    _member_classes  = list()
    _private_items = list()
    def __init__(self, data, parent=None):
        if isinstance(data,dict):
            pass
        elif isinstance(data, str):
            if valid_name(data):
                data = {'name':data}
            elif possible_obj_str(data,self):
                   parsed = lkml.load(data)
                   typ = self._type()+'s'
                   if typ in parsed.keys():
                       if len(parsed[typ]) == 1:
                           data = parsed[typ][0]
                       else:
                            raise Exception("Input string contains more than one " + self._type())
                   else:
                        raise Exception(f"Input string does not contain {typ}")
            else:
                raise Exception(
                                f"{data} does not match a " 
                                f"view: name {{}} or lookml name pattern")
        self.parent = parent
        self + data
    def __setattr__(self, key, value):
        if key in self._allowed_children():
            self._insert(key,value)
        else:
            object.__setattr__(self, key, value)

    def __repr__(self):
        return f'<{self.__class__.__name__}: {self.name} @{hex(id(self))}>'

    def __getattr__(self,key):
        if key.startswith('_add_hook'):
            if key in self.__dict__.keys():
                self.__dict__.__getitem__(key)
            else:
                None
        if key in self.__dict__.keys():
            self.__dict__.__getitem__(key)
        elif key in self._allowed_children():
            return prop_router(key,'__default__', self)
    
    def __getitem__(self,item):
        return self.__dict__.__getitem__(item)

    def __contains__(self,item):
        return item in self.__dict__

    def _allowed_children(self): return props._allowed_children[self._type()]
    def _conf(self): 
        if self._type() == 'model':
            return {
                'docs_url': 'https://docs.looker.com/reference/model-reference'
            }
        elif self._type() == 'manifest':
            return {
                'docs_url': 'https://docs.looker.com/reference/manifest-reference'
            }
        else:
            # return props.cfg[self.parent._type()][self._type()]
            return props.cfg[self._type()][self.parent._type()]

    def _dense(self): return len(list(self())) <= ws.dense_children_threshold 
    #P2: this causes a stack overflow: or self._length() <= ws.dense_str_len

    def __sub__(self,other):
        if isinstance(other, str):
            if other in self.__dict__.keys():
                del self.__dict__[other]
            else:
                warnings.warn(f'{other} did not exist on {self.name}')
        else:
            raise OperationError(self,' - ',other)

    def __iter__(self):
        def i():
            public_items = [
                v for k,v in 
                self.__dict__.items() 
                if k not in self._private_items
                ]
            for v in public_items:
                yield v
        self._valueiterator = iter(i())
        return self

    def __next__(self):
        try:
            return next(self._valueiterator)
        except:
            raise StopIteration

    def __call__(self, 
        type=prop, 
        exclude_type=tuple(), 
        sub_type=None, 
        exclude_subtype=None,
        ) -> Generator:
        for item in self:
            if isinstance(item,type)\
                and not isinstance(item, exclude_type)\
                and (item._type() in self._allowed_children())\
                and (True if not sub_type else item._type() == sub_type)\
                and (True if not exclude_subtype else item._type() not in exclude_subtype):
                yield item

    def _s(self, *args,**kwargs): 
        data = list(self(*args,**kwargs))
        string = ''
        for item in data:
            string += str(item)
        string = string + ' ' if len(data) > 0 else string
        # string = string + ws.nl if len(data) > 0 else string
        return string
    def _type(self) -> str: return self.__class__.__name__.lower()
    def __len__(self) -> int: return len([i for i in self()])
    def _length(self) -> int: return len([str(i) for i in self()])
    def __bool__(self) -> bool: return len(self) > 0
    def _json(self) -> dict: return lkml.load(str(self))
    def _insert(self,key: str, value: dict) -> None:
        #special top level namespace insertion for 'member classes', careful of filters keyword issues
        ac = self._allowed_children()
        if key in self._member_classes and self._type() not in props.cfg['filters'].keys():
            for member in value:
                self.__dict__.update(
                    { member['name'] : classMap[key](member,self) }
                )
        #normal property insertion
        elif key in ac:
            #step 1) instantiate the property
            candidate = prop_router( key, value, self )
            #step 2) if the hook lookup fails do nothing
            if self._add_hook(key, candidate):
                if isinstance(candidate,prop_named_construct):
                    if key not in self.__dict__.keys():
                        self.__dict__.update( { key: candidate } )
                    else:
                        self.__dict__[key].children.update(candidate.children)
                else:
                    self.__dict__.update( { key: candidate } )
            
        #special private attribute insertion, currently limited to 'name'
        elif key == 'name':
            self.__dict__.update( { key: value } )
        #else, skip and throw an invalid property warning
        else:
            warnings.warn(
                f'{key} skipped. {key} not a valid attribute of {self._type()} '
                f'refer to: {self._conf()["docs_url"]}'
                )
    def _add_hook_type(self,candidate_prop) -> bool:
        #currently doesn't do anything
        # candidate_prop
        return True
    ##### Legacy Methods #####
    def setProperty(self,key: str, value: dict):
        self._insert(key,value)
        return self

    def setName(self, name: str):
        """
        Set your Field's name value

        Args:
            name (str): a valid lookml name a to z, numbers and underscores

        Returns:
            Field: Your field for method chaining

        Raises:
            Exception: invalid lookml name
        """
        if valid_name(name):
            old_name = self.name
            self.name = name
            self.parent.__dict__.pop(old_name)
            self.parent + self
            return self
        else:
            raise Exception("invalid LookML Name")
    def setDescription(self, description: str):
        """
        Set your object's description

        Warning: 
        not all LookML objects support a description attribute. This method is present
        in a small number of cases on objects which do not support the parameter.

        Args:
            description (str): a description string

        Returns:
            Field: Your field for method chaining
        """
        self.description = description
        return self
    def hasProp(self, p: str) -> bool:
        """
            Check if a property is in your object

        Args:
            p (str): the string of the property
            i.e. dim.hasProp('type') -> True

        Returns:
            Boolean: True if prop present, false if not
        """
        if p in self.__dict__.keys():
            if isinstance(self.__dict__[p],prop):
                return True
            else:
                return False
        else:
            return False
    ##### Legacy Methods #####
    def _add_hook(self, key, candidate_val):
        callback = '_add_hook_' + key
        func = getattr(self, callback, None)
        if callable(func):
            return func(candidate_val)
        else:
            return True
    #P2: _add_hook_x is not invoked for name, want to add it so that valid_name can be run
    # def _add_hook_name(self,candidate):
    #     pass

    def __add__(self,other):
        # start object addition routine
        #P1: implment add hooks here and for member classes types
        if not isinstance(other,(str,dict)):
            if isinstance(self,View) and isinstance(other,Field):
                self.__dict__.update({other.name:other})

        # end object addition 
        if isinstance(other,str):
            tmp = lkml.load(other)
            if self._type() + 's' in tmp.keys():
                parsed = tmp[0][self._type() + 's']
            else:
                parsed = tmp
            # try:
            #     # Handles the special case where lkml passes single object wrapped in a collection list
            #     parsed = lkml.load(other)[self._type() + 's'][0]
            # except:
            #     parsed = lkml.load(other)

        elif isinstance(other,dict):
            parsed = other

        if isinstance(other,str) or isinstance(other,dict):
            for key,value in parsed.items():
                is_plural = True if (key[:-1] in props.plural_keys) else False
                key = key[:-1] if is_plural else key
                #special cases to treat as singular
                if self._type() in props.cfg['filters'].keys() and key == 'filter': 
                    key = key + 's'
                elif self._type() == 'explore_source' and key == 'bind_filter':
                    key = key + 's'
                elif self._type() == 'access_grant' and key == 'allowed_value':
                    key = key + 's'
                elif self._type() == 'query' and key in ("dimension","measure"):
                    key = key + 's'
                self._insert(key,value)
class Field(lookml):
    def __str__(self):
        if self._dense():
            return f'{ws.nl+(ws.s)}{self._type()}: {self.name} {{{ self._s() }}}'
        else:
            return (
                f'{ws.nl}{ws.s}{self._type()}: {self.name} {{'
                f'{ self._s(type=prop) }'
                f'{ws.nl}{ws.s}}}'
            )

    def __getattr__(self,key):
        #full reference
        if key == '__ref__' and self._type() in props.field_types:
            if self.parent is not None:
                return ('${' + self.parent.name + '.' + self.name + '}')
        #Short Reference
        elif key == '__refs__' and self._type() in props.field_types:
            return ('${' + self.name + '}')

        #full reference -- regex escaped
        elif key == '__refre__' and self._type() in props.field_types:
            if self.parent is not None:
                return (r'\$\{' + self.parent.name + r'\.' + self.name + r'\}')
        #Short reference -- regex escaped
        elif key == '__refsre__' and self._type() in props.field_types:
            if self.parent is not None:
                return (r'\$\{' + self.name + r'\}')
        #Raw Reference
        elif key == '__refr__' and self._type() in props.field_types:
            if self.parent is not None:
                return (self.parent.name + '.' + self.name)
        #Raw refence short
        elif key == '__refrs__' and self._type() in props.field_types:
            if self.parent is not None:
                return (self.name)
        #Raw Reference regex
        elif key == '__refrre__' and self._type() in props.field_types:
            if self.parent is not None:
                return (self.parent.name + r'\.' + self.name)
        #return prop defaults
        elif key in self._allowed_children():
            return prop_router(key,'__default__', self)
        else:
            super().__getattr__(key)
            #return prop_router(key,'__default__', self)
    
    def addTag(self,tag: str):
        """
        Add a tag to the field's tag collection

        Args:
            tag (str): the tag to add

        Returns:
            [type]: returns your field for method chaining

        """
        self.tags + tag
        return self
    def removeTag(self,tag: str):
        """
        Remove a tag from the field's tag collection

        Args:
            tag (str): the tag to remove

        Returns:
            [type]: returns your field for method chaining

        """
        self.tags - tag
        return self
    def setType(self,typ: str):
        """
        Sets the Fields type attribute to the specified value
        i.e. dim.setType('number')

        Args:
            typ (str): an allowed lookml type for the field

        Returns:
            Field: returns itself for chaining dim.setType('number').setPrimaryKey()
        """
        self.type = typ
        return self

    def setViewLabel(self, view_label: str):
        """
        sets the view_label attribute for the field
        controlling where it rolls up on the explore screen

        Args:
            view_label (str): view_label

        Returns:
            [type]: [description]
        """
        self.view_label = view_label
        return self

    def dependency_chain(self,i=0):
        """
        Yields nested tuple datastructures showing each of the
        dependency chains coming from your field
        
        Yields:
            Tuple: (1,dimA,(2,dimB,(3,dimC,...)))
        """
        if self.parent is not None:
            i += 1
            for dependent in list(self.parent.search('sql',[self.__refsre__,self.__refre__])):
                if dependent is not None:
                    for dep in dependent.dependency_chain(i=i):
                        if dep is not None:
                            yield  i, dependent, dep
                    yield  i, dependent

    def children_all(self,i=0):
        """
        Returns the fields in the same view which directly or indirectly 
        reference your field. I.e. dim B references dim A and dim C 
        references dim B. This method returns B and C

        Yields:
            Field: fields dependent upon your field
        """
        if self.parent is not None:
            i += 1
            for dependent in list(self.parent.search('sql',[self.__refsre__,self.__refre__])):
                if dependent is not None:
                    for dep in dependent.dependency_chain(i=i):
                        if dep is not None:
                            yield dep
                    yield dependent

    def children(self):
        """
        Returns the fields in the same view which directly reference 
        the field in their sql parameter

        Yields:
            Field: fields directly referencing your field
        """
        if self.parent is not None:
            for dependent in list(self.parent.search('sql',[self.__refsre__,self.__refre__])):
                if dependent is not None:
                    yield dependent

    def _ancestor_chain(self,i=0):
        for f in re.finditer(ws.mustachePattern, self.sql.value):
            yield f.group(1)

    def ancestors(self):
        """
        return the fields directly referenced by your field

        Yields:
            Field: a generator of direct ancestors
        """
        for ref in self._ancestor_chain():
            yield self.parent[ref]
    
    def ancestors_all(self):
        for f in re.finditer(ws.mustachePattern, self.sql.value):
            yield f.group(1)

    def _render_dependency_map(self,depChain,i=0):
        # depChain = (1, 'c', (2, 'd', (3, 'e')))
        # a -> c -> d -> e
        length = len(depChain)
        if i == 0:
            print(self.name, end=" -> ")
        if length == 3:
            print(depChain[1].name,end=" -> ")
            self._render_dependency_map(depChain[2],i=1)
        elif length == 2:
            print(depChain[1].name)

    def setName_replace_references(self, newName: str):
        """
        Alias for Field.setName_safe() method
        """
        self.setName_safe(newName)
        return self

    def setName_safe(self, newName: str):
        """
        Change the name of the field and references to it in sql 
        (does not yet perform the same for HTML / Links / Drill Fields / Sets / Actions etc)

        Args:
            newName (str): the new name

        Returns:
            self for method chaining
        """
        #P2: complete checking all places for dependencies, html, links etc
        old = copy.deepcopy(self.name)
        oldrefsre = copy.deepcopy(self.__refsre__)
        oldrefre = copy.deepcopy(self.__refre__)
        self.setName(newName)
        for f in self.parent.search('sql',[oldrefsre,oldrefre]):
            f.sql = re.sub(oldrefsre, self.__refs__, str(f.sql.value))
            f.sql = re.sub(oldrefre, self.__ref__, str(f.sql.value))
        self.parent.removeField(old)
        self.parent + self
        return self

    def setSql(self,sql: str):
        """
        Set the value of your sql
        We reccomend you 
        use the preffered method of setting directly: dim.sql = "${TABLE}.id"

        Args:
            sql (str): string of your sql

        Returns:
            self
        """
        self.sql = sql
        return self

    def addLink(self,url: str,label: str,icon_url: str ='https://looker.com/favicon.ico'):
        """
        Add a link

        Args:
            url (str): string for your url (can contain liquid)
            label (str): string for your label (can contain liquid)
            icon_url (str): default is "https://looker.com/favicon.ico"

        Returns:
            self
        """
        self + f"""
        link: {{
            url: "{url}"
            label: "{label}"
            icon_url: "{icon_url}"
        }}
        """
        return self

    def setString(self):
        """
        Sets the Dimension to type string

        Returns:
            Dimension: returns your dimension for chaining 
            dim.setPrimaryKey().setType('number')
        """
        self.type = 'string'
        return self

class Dimension(Field):
    def setPrimaryKey(self):
        """
        Sets the Dimension as it's parent view's primary key 
        and adds primary_key: yes to the field

        Returns:
            Dimension: returns your dimension for chaining 
            dim.setPrimaryKey().setType('number')
        """
        self.primary_key = 'yes'
        return self
    def unSetPrimaryKey(self):
        """
        Unregisters the Field as it's parents pk and adds
        primary_key: no to the field

        Returns:
            Dimension: returns your dimension for chaining 
            dim.setPrimaryKey().setType('number')
        """
        self.primary_key = 'no'
        return self
    def setDBColumn(self, dbColumn: str,changeIdentifier: bool=True):
        """
        Sets changes your sql to refer to the new column name

        Args:
            dbColumn (str): [description]
            changeIdentifier (bool, optional): [description]. Defaults to True.

        Returns:
            Dimension: returns your dimension for chaining 
            dim.setPrimaryKey().setType('number')
        """
        self.sql = (f'${{TABLE}}.{DB_FIELD_DELIMITER_START}'
                    f'{dbColumn}{DB_FIELD_DELIMITER_END}')
        if changeIdentifier:
            self.setName_safe(lookCase(dbColumn)) 
        return self
    def setNumber(self):
        """
        Sets the Dimension to type number

        Returns:
            Dimension: returns your dimension for chaining 
            dim.setPrimaryKey().setType('number')
        """
        self.type = 'number'
        return self
        
    def _add_hook_primary_key(self, candidate_prop):
        if self.parent._View__pk is not None:
            if self.parent._View__pk.primary_key:
                if candidate_prop:
                    raise DuplicatePrimaryKey( self.parent, self)
                else:
                    self.parent._View__pk = self
                    #returns true to allow the dict.update
                    return True 
            else:
                self.parent._View__pk = self
                return True
        else:
            self.parent._View__pk = self
            #returns true to allow the dict.update
            return True 


class Measure(Field): pass
class Filter(Field): pass
class Parameter(Field): pass
class Dimension_Group(Field): pass

class Model(lookml):
    _member_classes = [
        #  'explore'
    ]
    class dotdict(dict):
        def __getattr__(self,item):
            if item in self.__dict__.keys():
                return self.__dict__[item]
            elif item in self.keys():
                return self.__getitem__(item)
            else:
                return dict.__getattr__(item)
        def __iter__(self):
            self._valueiterator = iter(list(self.values()))
            return self

        def __sub__(self,other):
            if isinstance(other, str):
                del self[other]
            else:
                raise OperationError(self,' - ',other)

        def __next__(self):
            try:
                return next(self._valueiterator)
            except:
                raise StopIteration

        def __str__(self):
            tmp = ws.nl
            for i in self.values():
                tmp += (ws.nl + str(i))
            return tmp

    def __init__(self,data, parent=None):
        self.explores = self.dotdict()
        self.views = self.dotdict()
        self.parent = parent
        self + data
    
    def __add__(self,data):
        if isinstance(data, View):
            #handle refinements of objects in same scope by merge __dict__
            if data._is_refinement and data.name in self.views.keys():
                    self.views[data.name].__dict__.update(data.__dict__)
            else:
                data.parent = self
                self.views.update(
                        {data.name:data}
                            )
        if isinstance(data, Explore):
            #handle refinements of objects in same scope by merge __dict__
            if data._is_refinement and data.name in self.explores.keys():
                    self.explores[data.name].__dict__.update(data.__dict__)
            else:
                data.parent = self
                self.explores.update(
                        {data.name:data}
                            )
        if isinstance(data,dict):
            if 'explores' in data.keys():
                for explore in data['explores']:
                    #if refinement and pre-existing, merge object, else normal
                    if explore['name'] in self.explores.keys() and explore['name'].startswith('+'):
                        # self.explores[explore['name']].__dict__.update(explore)
                        self.__add__(Explore(explore, parent=self))
                    else:
                        self.explores.update(
                                {explore['name']:Explore(explore, parent=self)}
                                    )
                del data['explores']
            if 'views' in data.keys():
                for view in data['views']:
                    #if refinement and pre-existing, merge object, else normal
                    if view['name'] in self.views.keys() and view['name'].startswith('+'):
                        # self.views[view['name']].__dict__.update(view)
                        self.__add__(View(view, parent=self))
                    else:
                        self.views.update(
                                {view['name']:View(view, parent=self)}
                                    )
                del data['views']
            super().__add__(data)
    
        if isinstance(data,str):
            parsed = lkml.load(data)
            self.__add__(parsed)

    def __getattr__(self,key):
        if key in self.__dict__.keys():
            self.__dict__.__getitem__(key)
        elif key in self._allowed_children():
            return prop_router(key,'__default__', self)

    def __str__(self):
        return (
            f'{ self._s(sub_type="connection")}'
            f'{ self._s(exclude_subtype="connection") }'
            f'{ str(self.views) }'
            f'{ str(self.explores) }'
        )

class Explore(lookml):
    _member_classes = [
    ]
    def __init__(self,data, parent=None):
        super().__init__(data,parent)
        self._is_refinement = True if self.name.startswith('+') else False
    def _dense(self): return False
    def __str__(self): 
        return (f'{self._type()}: {self.name} {{'
                f'{ self._s(sub_type="join") }'
                f'{ self._s(type=(prop),exclude_subtype="join") }'
                f'{ws.nl}}}')

class Manifest(lookml):
    _member_classes = []
    def __str__(self):
        return (
            f'{ self._s() }'
        )

class View(lookml):
    """
    LookML View Object
    construct with name, short string or lkml json
    
.. code-block:: python
    :linenos:

    #construct via name string
    myView = lookml.View('foo')
    myView.view_label = "made by name"
    print(myView)
    >>> view: foo {
           view_label: "made by name"
         }
    #construct via string of LookML
    myView = lookml.View('view: foo { view_label: "made by string" }')
    print(myView)
    >>> view: foo {
           view_label: "made by string"
         }
    #construct via json (in lkml format)
    myView = lookml.View({'name':'foo', 'view_label':'made by json'})
    print(myView)
    >>> view: foo {
           view_label: "made by json"
         }

    """
    _member_classes = [
         'dimension'
        ,'measure'
        ,'dimension_group'
        ,'filter'
        ,'parameter'
    ]
    _private_items = ['_View__pk']
    def __init__(self, data, parent=None):
        self.__pk = None
        super().__init__(data, parent)
        self._is_refinement = True if self.name.startswith('+') else False

    def __sub__(self,other):
        if isinstance(other, Field):
            if other.name in self.__dict__.keys():
                if len(list(other.dependency_chain())) > 1:
                    warnings.warn(
                        f'{other.__refrs__} had dependencies: '
                        f'{[f.name for f in other.children_all()]}'
                        )
                del self.__dict__[other.name]
            else:
                warnings.warn(f'{other.name} did not exist in {self.name}')
        elif isinstance(other, str):
            if other in self.__dict__.keys():
                if isinstance(self[other],Field):
                    if len(list(self[other].dependency_chain())) > 1:
                        warnings.warn(
                            f'{self[other].__refrs__} had dependencies: '
                            f'{[f.name for f in self[other].children_all()]}'
                            )
                del self.__dict__[other]
            else:
                warnings.warn(f'{other} did not exist on {self.name}')
        else:
            raise OperationError(self,' - ',other)
    def _fields(self) -> Generator: return self(type=Field)
    def _dims(self) -> Generator: return self(type=Dimension)
    def _dim_groups(self) -> Generator: return self(type=Dimension_Group)
    def _measures(self) -> Generator: return self(type=Measure)
    def _params(self) -> Generator: return self(type=Parameter)
    def _filters(self) -> Generator: return self(type=Filter)
    def _dense(self) -> Generator: return False

    def _add_hook_sql_table_name(self, sql_table_name):
        if 'derived_table' in self:
            raise CoexistanceError(
                self.derived_table,
                sql_table_name,
                additional_message='Try running del yourViewName.derived_table '
                +'before adding the sql_table_name'
                )
        else:
            return True
    def _add_hook_derived_table(self, derived_table):
        if 'sql_table_name' in self:
            raise CoexistanceError(
                self.sql_table_name,
                derived_table,
                additional_message='Try running del yourViewName.sql_table_name '
                +'before adding the derived_table'
                )
        else:
            return True

    def search(self, prop, pattern):
        '''
        pass a regex expression and will return the fields whose sql match
        :param prop: name of proprty you'd like to search
        :param pattern: the regex pattern
        :type prop: str
        :type patter: a regex search string
        :return: a generator / iteratble set of fields who have a member property matching the pattern
        :rtype: Field
        '''
        if isinstance(pattern,list):
            pattern = '('+'|'.join(pattern)+')'
        find_this = r''.join([r'.*',pattern,r'.*'])
        for field in self.fields():
            if prop in field.__dict__.keys():
                string_to_search = field.__dict__[prop].value
            else:
                string_to_search = ''
            if re.match(find_this,string_to_search):
                yield field

    def removeField(self,field):
        """
        Remove a Field from your View

        Args:
            field ([type]): [description]

        Returns:
            View: returns your view for method chaining 
            i.e. view.removeField(view.id).addCount()
        """
        self - field
        return self

    def __str__(self):
        return (
            f'{self._type()}: {self.name} {{'
                f'{ self._s(exclude_subtype="set") }'
                f'{ self._s(type=Filter) }'
                f'{ self._s(type=Parameter) }'
                f'{ self._s(type=Dimension) }'
                f'{ self._s(type=Dimension_Group) }'
                f'{ self._s(type=Measure) }'
                f'{ self._s(sub_type="set") }'
            f'{ws.nl}}}'
        )
    ##### Legacy Methods #####
    # generators #
    def fieldNames(self):
        """
        returns all the field names

        Yields:
            [type]: iteration of the fields names
        """
        for field in self._fields():
            yield field.name

    def fields(self): 
        """
        get all the Views Fields

        Returns:
            [type]: iterable collection of Fields
        """
        for f in self._fields():
            yield f

    def dimensions(self): 
        """
        get all the Views Dimensions

        Returns:
            Dimension: iterable collection of Dimensions
        """
        for d in self._dims():
            yield d

    def measures(self): 
        """
        get all the Views Measures

        Returns:
            Measure: iterable collection of Measures
        """
        for d in self._measures():
            yield d

    def dimensionGroups(self): 
        """
        iterate over the dimension_group fields in your view

        Returns:
            Iterator: Iterator of type dimension_group Field
        """
        for dg in self._dim_groups():
            yield dg

    def filters(self): 
        """
        iterate over the filter fields in your view

        Returns:
            Iterator: Iterator of type Filter Field
        """
        for f in self._filters():
            yield f 

    def parameters(self): 
        """
        iterate over the parameter fields in your view

        Returns:
            Iterator: Iterator of type Parameter Field
        """
        for p in self._params():
            yield p

    def getFieldsByTag(self,tag: str):
        """
        Loop through your fields with a specific tag present

        Args:
            tag (str): 
        Yields:
            Field: Fields with the selected tag
        """
        for f in self._fields():
            if tag in f.tags:
                yield f

    def getFieldsByType(self, typ: str):
        """
        Loop over fields with a certain type

        Args:
            typ (str): 'string','number','count'...

        Yields:
            Field: Field with the selected type attribute
        """
        for f in self._fields():
            if f.type.value == typ:
                yield f

    def getFieldsSorted(self):
        '''
        return the fields sorted first conventionally by type, then alphabetically

        Yields:
            Field: a list of fields in the natural sort order
        '''
        return sorted(self._fields(), key=lambda field: ''.join([str(isinstance(field, Measure)), field.name]))
    # end generators
        
    def addDimension(self,d: (str,Dimension),type: str = 'string'): 
        """
        Add a dimension object, or add a string DB column you would like added 

        Args:
            d ([type]): [description]
            type (str, optional): [description]. Defaults to 'string'.

        Returns:
            [type]: [description]
        """
        if isinstance(d,Dimension):
            self + d
        if isinstance(d,str):
            name = lookCase(d)
            self + f'''
            dimension: {name} {{
                type: {type}
                sql: {d} ;;
            }}
            '''
        return self

    def addCount(self): 
        """
        Adds a basic measure type count to your view

        Returns:
            self: allows for chaining myView.addCount().addDimension()
        """
        self + 'measure: count { type: count }'
        return self

    def addAverage(self, field: Dimension):
        """
        Add an average of a dimension to your view

        Args:
            field (Dimension): add a dimension object you'd like the average on
        """
        self + f'''
            measure: {field.name}_avg {{
                type: average
                sql: {field.__ref__} ;;
            }} 
        '''
        return self

    def addCountDistinct(self,field: Dimension): 
        """
        Add a count distinct of a dimension to your view

        Args:
            field (Dimension): add a dimension object you'd like the distinct on
        """
        self + f'''
            measure: {field.name}_count_distinct {{
                type: count_distinct
                sql: {field.__ref__} ;;
            }} 
        '''
        return self

    def sum(self,field: Dimension): 
        """
        Add a sum of a dimension to your view

        Args:
            field (Dimension): add a dimension object you'd like the sum on
        """
        self + f'''
            measure: {field.name}_sum {{
                type: tum
                sql: {field.__ref__} ;;
            }} 
        '''
        return self

    def sumAllNumDimensions(self):
        """
        creates a sum measure for each Dimension of type number

        Returns:
            View: returns your View for method chaining view.sumAllNumDimensions().addCount()...
        """
        for numDim in self.getFieldsByType('number'):
            self.sum(numDim)
        return self

    def setViewLabel(self,label: str):
        """
        Add a view_label to your view

        Args:
            label (str): the text of the View Label

        Returns:
            View: returns your View for method chaining view.sumAllNumDimensions().addCount()...
        """
        self.setProperty('view_label',label)
        return self

    def print_dependency_map(self):
        for field in list(self.fields()):
            for referenceChain in field.dependency_chain():
                field._render_dependency_map(referenceChain)

    def first_order_fields(self) -> Generator:
        '''
        generates fields that reference DB fields directly
        having the ${TABLE} syntax

        :return: Generator of type Field
        :rtype: Field
        '''
        for field in self(type=Field):
            if '${TABLE}.' in field.sql:
                yield field
    ##### Legacy Methods #####

class prop_block(prop):
    def __str__(self):
        dense = True if self.parent._dense() else False
        __ = ws.nl + (ws.s * self.conf['indent']) if not dense else ws.s
        return f'{__}{self.key}: {self.value} ;;'

class prop_string(prop):
    def __str__(self):
        dense = True if self.parent._dense() else False
        __ = ws.nl + (ws.s * self.conf['indent']) if not dense else ws.s
        return f'{__}{self.key}: "{self.value}"'
class prop_string_unquoted(prop): pass

class prop_named_construct_single(Field):
    def __init__(self,key, data, parent, conf={}):
        self._key = key
        self.conf = conf
        self + data
    def _type(self): return self._key
    def __getattr__(self,key):
        if key in self.__dict__.keys():
            self.__dict__.__getitem__(key)
    def __setattr__(self, key, value):
        #P2: way to get arround this setattr manual exceptions?
        if key in ('_key','conf'):
            object.__setattr__(self, key, value)
        elif key in props._allowed_children[self._type()]:
            self.__dict__.update( { key : prop_router( key, value, self) } )
        else:
            object.__setattr__(self, key, value)

    def __str__(self):
        i = self.conf['indent']
        if self._dense():
            return f'{ws.nl+(ws.s*i)}{self._type()}: {self.name} {{{ self._s() }}}'
        else:
            return (
                f'{ws.nl}{ws.s*i}{self._type()}: {self.name} {{'
                f'{ self._s(type=prop) }'
                f'{ws.nl}{ws.s*i}}}'
            )
class prop_named_construct(prop):
    def __init__(self, key, value, parent, conf={}):
        self.key = key
        self.value = value
        self.parent = parent
        self.conf = conf
        self.children = {}
        for instance in value:
            child = prop_named_construct_single(key, instance, self, conf=props.cfg[key][parent._type()])
            self.children.update({child.name:child})

    def __setattr__(self, key, value):
        if key in ('key','value','parent','conf','children'):
            object.__setattr__(self, key, value)
        if key in props._allowed_children[self.key]:
            self.__dict__.update( { key : prop_router( key, value, self) } )
        else:
            object.__setattr__(self, key, value)

    def __getattr__(self,key):
        if key in self.children.keys():
            return self.children.__getitem__(key)
        elif key in self.__dict__.keys():
            return self.__dict__.__getitem__(key)
        else:
            return None

    def __getitem__(self,item):
        # if isinstance(item, int):
        return self.children[item]
        
    
    def __add__(self,other):
        #P3: since this is on the props hierarchy, need to refactor addition so it's not challenging
        #do after extensive unit tests are in place
        if isinstance(other,str):
            try:
                parsed = lkml.load(other)[self.key+'s']
            except:
                raise Exception(f'can only add {self.key} to {self.key}')

        elif isinstance(other,dict):
            parsed = other
        for item in parsed:
            candidate = prop_named_construct_single(
                    self.key, 
                    item, 
                    self.parent, 
                    conf=props.cfg[self.key][self.parent._type()]
                    )
            self.children.update({candidate.name: candidate})

    def __str__(self):
        dense = True if self.parent._dense() else False
        __ = ws.nl + (ws.s * self.conf['indent']) if not dense else ws.s
        rendered = ''
        # for name, child in self.children.items():
        for child in self.children.values():
            rendered += f'''{ child }'''
        return rendered

    def __sub__(self,other: str):
        if other in self.__dict__.keys():
            del self.__dict__[other]
        elif other in self.children.keys():
            del self.children[other]
        else:
            raise Exception(f'{other} not present in {self.key}')

    def __contains__(self,other):
        return other in self.children
    def __iter__(self):
        self._valueiterator = iter(list(self.children.values()))
        return self

    def __next__(self):
        try:
            return next(self._valueiterator)
        except:
            raise StopIteration



class prop_anonymous_construct(prop):
    def __init__(self, key, value, parent, conf={}):
        self.key = key
        self.value = value
        self.parent = parent
        self.conf = conf
        self.children = {}
        for name,child in value.items():
            self.children.update({name: prop_router(name,child,self)})

    def __setattr__(self, key, value):
        #P2: fix this, I think it can be done without hard coding the variable names
        if key in ('key','value','parent','conf','children'):
            object.__setattr__(self, key, value)
        if key in props._allowed_children[self.key]:
            self.__dict__.update( { key : prop_router( key, value, self) } )
        else:
            object.__setattr__(self, key, value)

    def __getattr__(self,key):
        if key in self.__dict__.keys():
            self.__dict__.__getitem__(key)
        elif key in self.children.keys():
            return self.children[key]
        else:
            return None

    def __sub__(self,other: str):
        if other in self.__dict__.keys():
            del self.__dict__[other]
        elif other in self.children.keys():
            del self.children[other]

#P1: does not support addition like proper lookml objects, 
# pull out add and add_hook, insert, sub and sub_hook systems into a  
# common container class

    def print_children(self):
        rendered_children = ''
        for child in self.children.values():
            rendered_children += (child.conf['indent']*ws.s) + str(child)
        rendered_children += ws.nl + (self.conf['indent'] * ws.s)
        return rendered_children
    def _dense(self): return len(self.children) == 0

    def __str__(self):
        dense = True if self.parent._dense() and self._dense() else False
        i = self.conf['indent']
        __ = ws.nl + (ws.s * i) if not dense else ws.s
        return f'''{__}{self.key}: {{ { self.print_children() } }}'''

class prop_anonymous_construct_plural(prop):
    class prop_anonymous_construct_child(prop_anonymous_construct):
        def remove(self):
            """
                Remove self from construct
            """
            self.parent.link.children.remove(self)

    def __init__(self, key, value, parent, conf={}):
        self.key = key
        self.value = value
        self.parent = parent
        self.conf = conf
        self.children = []
        for construct in value:
            child = self.prop_anonymous_construct_child(
                        key, 
                        construct, 
                        self.parent, 
                        conf=conf
                        )
            self.children.append(child)

    def remove(self,item):
        """
        remove an item from your 

        Args:
            item ([type]): [description]
        """
        self.children.remove(item)
    
        
    def __setattr__(self, key, value):
        if key in ('key','value','parent','conf','children'):
            object.__setattr__(self, key, value)
        elif self.key != 'filters':
            if key in props._allowed_children[self.key]:
                self.__dict__.update( { key : prop_router( key, value, self) } )
        else:
            object.__setattr__(self, key, value)

    def __getattr__(self,key):
        if key in self.__dict__.keys():
            self.__dict__.__getitem__(key)
        else:
            return None
    def __getitem__(self,item):
        if isinstance(item, int):
            return self.children[item]
    
    def __add__(self,other):
        #P3: since this is on the props hierarchy, need to refactor addition so it's not challenging
        #do after extensive unit tests are in place
        if isinstance(other,str):
            try:
                parsed = lkml.load(other)[self.key+'s']
            except:
                raise Exception(f'can only add {self.key} to {self.key}')

        elif isinstance(other,dict):
            parsed = other
        for item in parsed:
            candidate = self.prop_anonymous_construct_child(
                    self.key, 
                    item, 
                    self.parent,
                    conf=props.cfg[self.key][self.parent._type()]
                    )
            self.children.append(candidate)

    def __str__(self):
        dense = True if self.parent._dense() and self._dense() else False
        __ = ws.nl + (ws.s * self.conf['indent']) if not dense else ws.s
        rendered = ''
        for child in self.children:
            rendered += f'''{ str(child) }'''
        return rendered

class common_list_functions(object):
    def __add__(self, other):
        if isinstance(other, str):
            self.value.append(other)
        elif isinstance(other, list):
            for item in other:
                self.value.append(item)
    def __sub__(self,other):
        if isinstance(other, str):
            self.value.remove(other)
        elif isinstance(other, list):
            for item in other:
                self.value.remove(item)
    def __contains__(self,other):
        return other in self.value
    def __iter__(self):
        self._valueiterator = iter(self.value)
        return self

    def __next__(self):
        try:
            return next(self._valueiterator)
        except:
            raise StopIteration

class prop_list_unquoted(prop, common_list_functions): 
    def __str__(self):
        dense = True if self.parent._dense() else False
        i = self.conf['indent']
        __ = ws.nl + (ws.s * i) if not dense else ws.s
        if len(self.value) > ws.list_multiline_threshold:
            return (f'{__}{self.key}: { "[" +ws.nl + (ws.s*(i+1))}'
                    f'{(", " + ws.nl + (ws.s*(i+1))).join(self.value) }'
                    f'{ ws.nl + (ws.s*(i+1)) + "]" }'
                    )
        else:
            return f'{__}{self.key}: { "[" + ", ".join(self.value) + "]" }'

class prop_list_quoted(prop, common_list_functions):
    def __str__(self):
        dense = True if self.parent._dense() else False
        i = self.conf['indent']
        __ = ws.nl + (ws.s * i) if not dense else ws.s

        if len(self.value) > ws.list_multiline_threshold:
            return (f'''{__}{self.key}: { '[' + ws.nl + (ws.s*(i+1)) + '"'}'''
                    f'''{('", ' + ws.nl + (ws.s*(i+1)) + '"').join(self.value) }'''
                    f'''{ '"' + ws.nl + (ws.s*(i+1)) + ']' }'''
                    )
        elif len(self.value) > 0:
            return f'''{__}{self.key}: { '["' + '","'.join(self.value) + '"]' }'''
        else:
            return f'{__}{self.key}: { "[]" }'

class prop_includes(prop, common_list_functions):
    def __str__(self):
        rendered = [f'include: "{val}"' for val in self.value]
        return ws.nl.join(rendered)

class prop_options(prop):
    def __init__(self, key, value, parent, conf={}):
        self.key = key
        self.value = value
        self.parent = parent
        self.conf = conf
        assert value in conf['allowed_values']

class prop_options_quoted(prop_string): 
    def __init__(self, key, value, parent, conf={}):
        self.key = key
        self.value = value
        self.conf = conf
        assert value in conf['allowed_values']

class prop_expression(prop): pass
class flt(object):
    def __init__(self,key,value):
        self.key, self.value = key, value
    
    def __str__(self):
        return f'{self.key}:"{self.value}"'
    
    def contains(self,expression):
        self.value = '%' + expression + '%'

class srt(object):
    #sorts: [field_name_1: asc | desc, field_name_2: asc | desc, …]
    def __init__(self,key,value):
        self.key, self.value = key, value
    
    def __str__(self):
        return f'{self.key}: {self.value} '
    
    def asc(self):
        self.value = 'asc'
    def desc(self):
        self.value = 'desc'

class prop_filters(prop_anonymous_construct_plural): 
    # Old Style:
    #{'filters': [{'field': 'foo', 'value': '%cool%'}, {'field': 'bar', 'value': '%great%'}], 'name': 'filters'}
    # New Style:
    #{'filters': {'foo': '%cool%', 'bar': '%great%'}, 'name': 'filters'}
    def __init__(self, key, value, parent, conf={}):
        self.key = key
        self.parent = parent
        self.conf = conf
        if isinstance(value,list):
            self.value = {f['field']:flt(f['field'],f['value']) for f in value}
        else:
            self.value = {k:flt(k,v) for k,v in value.items()}

    def __str__(self):
        fltrs = [ str(f)  for f in self.value.values()]
        indent = ws.nl + (ws.s * self.conf['indent']) if not self.parent._dense() else ws.s
        if len(self.value) > ws.list_multiline_threshold:
            return (f'{indent}{self.key}: [ {ws.nl}       ' 
                    f"{(chr(10)+'      ,').join(fltrs)}"
                    '{ws.nl}      ] ')
        else:
            return f"{indent}{self.key}: [ {', '.join(fltrs)} ] "

    def __getattr__(self,key):
        if key in self.__dict__.keys():
            return self.__dict__[key]
        elif key in self.value.keys():
            return self.value[key]

    def __getitem__(self,item):
        self.value[item]

    def __add__(self,other):
        if isinstance(other,dict):
            other = {k:flt(k,v) for k,v in other.items()}
            self.value.update(other)
    
    def __sub__(self,other):
        if isinstance(other,str):
            del self.value[other]

class prop_sorts(prop_filters):
    def __init__(self, key, value, parent, conf={}):
        self.key = key
        self.parent = parent
        self.conf = conf
        if isinstance(value,list):
            self.value = {f['field']:srt(f['field'],f['value']) for f in value}
        else:
            self.value = {k:srt(k,v) for k,v in value.items()}

class prop_yesno(prop):
    def __bool__(self): return True if self.value == 'yes' else False

class prop_html(prop_block): pass
class prop_sql(prop_block):
    def nvl(self):
        self.value = f"nvl({self.value},0)"

def prop_router(key,value, parent):
    is_plural = True if key[:-1] in props.plural_keys else False
    key = key[:-1] if is_plural else key
    parent_type = parent._type()
    #P1: not proud of this, but the "plural keys" needs to be corrected for certain contexts. This works for now but I can't explain it well
    if parent_type in props.cfg['filters'].keys() and key == 'filter':
        key = key + 's'
    elif parent_type == 'explore_source' and key == 'bind_filter':
        key = key + 's'
    elif parent_type == 'query' and key in ('dimension','measure'):
        key = key + 's'
    elif parent_type == 'access_grant' and key == 'allowed_value':
        key = key + 's'
    if key == 'query' and parent_type == 'aggregate_table':
        #P1: make more elegant / clean up this fixes an issue where because query is a named construct in the explore
        # and an anonymous construct in the aggregate table I need to take the 1 and only value out because lkml thinks
        # it's a plural key after adding it to the plural keys to fix the issue with explore > query
        value = value[0]
    conf = props.cfg[key][parent_type]
    prop_type = conf['type']
    default = conf['default_value']
    prop_map = {
         'anonymous_construct_plural': prop_anonymous_construct_plural
        ,'named_construct': prop_named_construct
        ,'named_construct_single': prop_named_construct_single
        ,'list_unquoted': prop_list_unquoted
        ,'yesno': prop_yesno
        ,'list_quoted': prop_list_quoted
        ,'string_unquoted': prop_string_unquoted
        ,'anonymous_construct': prop_anonymous_construct
        ,'string': prop_string
        ,'options': prop_options
        ,'options_quoted': prop_options_quoted
        ,'expression': prop_expression
        ,'filters': prop_filters
        ,'html': prop_html
        ,'sorts': prop_sorts
        ,'sql': prop_sql
        ,'include': prop_includes
    }
    if value == '__default__':
        return prop_map[prop_type]( key, default, parent, conf=conf)
    else:
        if key in classMap.keys():
            return None
        else:
            return prop_map[prop_type]( key, value, parent, conf=conf)

classMap = {
     'view': View
    ,'dimension': Dimension
    ,'measure': Measure
    ,'parameter': Parameter
    ,'filter': Filter
    ,'dimension_group': Dimension_Group
    ,'explore': Explore
    ,'model': Model
    ,'manifest': Manifest
}




import src.lookml.lookml as lookml
import src.lookml.lkml as lkml

import unittest, copy
from pprint import pprint
import configparser, json
# from looker_sdk import client, models, methods
from looker_sdk import models, methods, init31
config = configparser.ConfigParser()
config.read('tests/settings.ini')

class testScenarioEAV(unittest.TestCase):
    '''
        Objective: test end to end common use case scenarios 
    '''
    def setUp(self):
        pass
    def test_eav_unnester(self):
        sdk = init31("api.ini")
        sql_for_fields = f"""
                SELECT 
                     cpf.org_id
                    ,cpf.value
                    ,cpf.datatype
                    ,cpf.field_name as "FIELD_NAME"
                    , CASE 
                        WHEN cpf.datatype IN ('TIMESTAMP_LTZ') THEN 'time'
                        WHEN cpf.datatype IN ('FLOAT','NUMBER', 'int') THEN 'number'
                        ELSE 'string' END as "LOOKER_TYPE"
                FROM 
                    -- public.custom_profile_fields as cpf
                    (
                        SELECT 1 as user_id, 8 as org_id, 'c_donation_amount' as field_name, '40' as value, 'int' as datatype UNION ALL
                        SELECT 1, 8, 'c_highest_achievement', 'gold badge', 'varchar' UNION ALL
                        SELECT 2, 101, 'c_highest_achievement', 'silver badge', 'varchar' UNION ALL
                        SELECT 2, 101, 'c_monthly_contribution', '300', 'int' UNION ALL
                        SELECT 3, 101, 'c_highest_achievement', 'bronze badge', 'varchar' UNION ALL
                        SELECT 3, 101, 'c_monthly_contribution', '350', 'int' UNION ALL
                        SELECT 4, 101, 'c_monthly_contribution', '350', 'int' UNION ALL
                        SELECT 4, 101, 'age', '32', 'int' UNION ALL
                        SELECT 5, 102, 'c_monthly_contribution', '100', 'int'
                    ) as cpf
                WHERE
                    1=1
                GROUP BY 1,2,3,4,5
                LIMIT 100
        """
        query_config = models.WriteSqlQueryCreate(sql=sql_for_fields, connection_id="snowlooker")
        query = sdk.create_sql_query(query_config)
        response = json.loads(sdk.run_sql_query(slug=query.slug, result_format="json"))

        proj = lookml.Project(
                 repo= config['github']['repo']
                ,access_token=config['github']['access_token']
                ,looker_host="https://profservices.dev.looker.com/"
                ,looker_project_name="russ_sanbox"
                ,branch='dev-russell-garner-b7gj'
        )

        modelFile = proj['eav_example/eav.model.lkml']

        eavSource = modelFile['views']['eav_source']
        flatteningNDT = modelFile['views']['usr_profile']

        #Step 0) Ensure there is a hidden explore to expose the eav_souce for NDT
        modelFile + f'''
            explore: _eav_flattener {{
                from: {eavSource.name}
                hidden: yes
            }}
        '''
        #Begin the derived table, will be added to as we loop through the fields
        drivedtableString = f'''
            derived_table: {{
                explore_source: _eav_flattener {{
                    column: user_id {{ field: _eav_flattener.user_id }}
                    column: org_id {{ field: _eav_flattener.org_id }}
        '''

        #Set up a list to track the unique org ids and column names
        orgIds, columns = [], []

        for column in response:
            dimName = lookml.lookCase(column['FIELD_NAME'])
            orgIds.append(column['ORG_ID'])
            columns.append(dimName)
            #Step 1) Add flattening measure to the EAV source table
            eavSource + f'''
                  measure: {dimName} {{
                        type: max
                        sql: CASE WHEN ${{field_name}} = '{column['FIELD_NAME']}' THEN ${{value}} ELSE NULL END;;
                    }}
            '''

            #Step 2) Add to the NDT fields
            flatteningNDT + f'''
                    dimension: {dimName}_org_{column['ORG_ID']} {{
                        label: "{dimName}"
                        type: {column['LOOKER_TYPE']}
                        sql: ${{TABLE}}.{dimName} ;;
                        required_access_grants: [org_{column['ORG_ID']}]
                    }}
            '''
            if column['LOOKER_TYPE'] == "number":
                flatteningNDT + f'''
                    measure: {dimName}_total_org_{column['ORG_ID']} {{
                        label: "{dimName}_total"
                        type: sum
                        sql: ${{{dimName}_org_{column['ORG_ID']}}} ;;
                        required_access_grants: [org_{column['ORG_ID']}]
                    }}
                '''
        #Finalize / add the completed derived table string to the NDT:
        for col in set(columns):
            drivedtableString += f' column: {col} {{ field: _eav_flattener.{col} }}'
        drivedtableString += '}}'

        #Create the access grants for each org
        accessGrants = ''
        for org in set(orgIds):
            accessGrants += f'''
                access_grant: org_{org} {{
                user_attribute: org_id
                allowed_values: [
                    "{org}"
                ]
                }}
            '''

        #Finish by adding the derived table to the 
        flatteningNDT + drivedtableString
        modelFile + accessGrants

        proj.put(modelFile)
        # proj.deploy()



 



    


import looker_sdk
import argparse
import os
import re
import requests
import json
import os.path
from os import path

#parse inputs
arguments = argparse.ArgumentParser()
arguments.add_argument('--dev-base-url',type=str)
arguments.add_argument('--dev-client-id',type=str)
arguments.add_argument('--dev-client-secret',type=str)
arguments.add_argument('--googledemo-client-id',type=str)
arguments.add_argument('--googledemo-client-secret',type=str)
arguments.add_argument('--project-name',type=str)
arguments.add_argument('--repo-path',type=str)

args = arguments.parse_args()

googledemo_base_url = 'https://googledemo.looker.com'
project_name = args.project_name
repo_path = args.repo_path

#create sdk connection to googledemo in order to find the dashboards 
os.environ['LOOKERSDK_BASE_URL']=googledemo_base_url+str(':19999')
os.environ['LOOKERSDK_CLIENT_ID']=args.googledemo_client_id
os.environ['LOOKERSDK_CLIENT_SECRET']=args.googledemo_client_secret
sdk = looker_sdk.init31()

#get the dashboards
dashboards = json.loads(sdk.run_look('44',result_format='json'))
dashboards_dict = {}
for dash in dashboards:
    if dash['core_demos.lookml_project_name'] == project_name:
        dashboards_dict[dash['demo_dashboards.development_dashboard_id']] = dash
dashboard_ids = dashboards_dict.keys()
sdk.logout()

#run content validator on dev
os.environ['LOOKERSDK_BASE_URL']=args.dev_base_url+str(':19999')
os.environ['LOOKERSDK_CLIENT_ID']=args.dev_client_id
os.environ['LOOKERSDK_CLIENT_SECRET']=args.dev_client_secret
sdk = looker_sdk.init31()

print('Running Content Validation')
results = sdk.content_validation()

broken_dashboards = [content.dashboard.id for content in results.content_with_errors if content.dashboard and content.dashboard.id in dashboard_ids]
with open(os.path.join(repo_path,'.github',"dashboard_errors.txt"),"w") as f:
    for i,dash in enumerate(broken_dashboards):
        if dash != broken_dashboards[0]:
            if i > 0:
                f.write(', ')
            f.write('<{} | {}>'.format(args.dev_base_url+'/dashboards/'+str(dash_id), dashboard_title))
        #f.write('\n' + args.dev_base_url +'/dashboards/'+str(dash))
    f.close()

def prepend_line(file_name, line):
    """ Insert given string as a new line at the beginning of a file """
    # define name of temporary dummy file
    dummy_file = file_name + '.bak'
    # open original file in read mode and dummy file in write mode
    with open(file_name, 'r') as read_obj, open(dummy_file, 'w') as write_obj:
        # Write given line to the dummy file
        write_obj.write(line + '\n')
        # Read lines from original file one by one and append them to the dummy file
        for line in read_obj:
            write_obj.write(line)
    # remove original file
    os.remove(file_name)
    # Rename dummy file as the original file
    os.rename(dummy_file, file_name)

# for each dashboard get the lookml 
dash_to_import = []
#erase contents of the dashboard
open(os.path.join(repo_path,'.github',"new_dashboards.txt"),"w").close()
blank = 0
dash_path = os.path.join(repo_path,'dashboards')
if not path.exists(dash_path):
    os.mkdir(dash_path)
    #add include on all model files
    for root, dirs, files in os.walk(repo_path):
        files = [f for f in files if not f[0] == '.']
        dirs[:] = [d for d in dirs if not d[0] == '.']
        for file in files:
            if ".model" in file:
                if root == repo_path:
                    prepend_line(os.path.join(root, file), 'include: "dashboards/*.lookml"')
                else:
                    prepend_line(os.path.join(root, file), 'include: "/dashboards/*.lookml"')

def get_dashboard_path(filename):
    #check to see if dashboard exists in subfolder in /dashboards
    path = os.path.join(dash_path,filename)
    for root, dirs, files in os.walk(dash_path):
        for file in files:
            if filename == file:
                return os.path.join(root, file)
    return path
    

for dash_id in list(dashboard_ids):
    s_dash = sdk.dashboard(str(dash_id))
    dashboard_title = s_dash.title
    #dont import broken content
    if dash_id not in broken_dashboards:
        #check if the dashboard lookml file already exists
        # d_filename = re.sub('[^a-zA-Z0-9 \n\.]', '_', dashboard_title).replace(' ','_').lower()+'.dashboard.lkml'
        d_filename = dashboards_dict[dash_id]['demo_dashboards.lookml_dashboard_id'].split('::')[1]+'.dashboard.lookml'
        # dashboards_dict[dash_id]['filename'] = d_filename
        d_file_path = get_dashboard_path(d_filename)
        if os.path.exists(d_file_path):
            print(str(dash_id), ' dashboard already exists, reading lookml')
            with open(d_file_path,"r") as f:
                existing_dash = f.read()
        else:
            existing_dash = ''
            #if it does not exist add it to new dashboards text
            with open(os.path.join(repo_path,'.github',"new_dashboards.txt"),"a") as f:
                # f.write('\n' + dashboard_title + ': ' + args.dev_base_url+'/dashboards/'+str(dash_id))
                if blank != 0:
                    f.write(', ')
                f.write('<{} | {}>'.format(args.dev_base_url+'/dashboards/'+str(dash_id), dashboard_title))
                blank +=1
        
        new_dash = sdk.dashboard_lookml(str(dash_id)).lookml
        if existing_dash != new_dash:
            print(str(dash_id) + ' dashboard doesnt exist or has changed, lookml is being updated')
            dash_to_import.append(dash_id)
            #overwrite file with new contents
            with open(d_file_path,"w+") as f:
                f.seek(0)
                f.write(new_dash)
                f.truncate()
sdk.logout() 

                



import unittest
import lookml

class testTopN(unittest.TestCase):
    def setUp(self):
        pass
    def test_topN(self):
        def apply_top_n(
            project, 
            view_file, 
            view, 
            rank_by_dims, 
            rank_by_meas, 
            model_file, 
            explore, 
            agg_view='rank_ndt', 
            dynamic_dim_name='dynamic_dim', 
            dynamic_dim_selector='dynamic_dim_selector'):
            #### SETUP ####
            p = project
            mf = p[model_file]
            vf = p[view_file]
            v = vf['views'][view]
            e = mf['explores'][explore]
            #### DO WORK ####
            #Add the parameter to the initial view file
            dynamic_dim_sql = ''
            i = 0
            for key,val in rank_by_dims.items():
                if i == 0:
                    dynamic_dim_sql = f"{{% if {v.name}.{dynamic_dim_selector}._parameter_value == '{key}' %}} {val}"
                else:
                    dynamic_dim_sql = dynamic_dim_sql + '\n' + f"{{% elsif {v.name}.{dynamic_dim_selector}._parameter_value == '{key}' %}} {val}"
                i = 1 + 1
            dynamic_dim_sql = dynamic_dim_sql + f"""
                    {{% else %}} 'N/A' 
                    {{% endif %}}
                """
            allowed_values = ''
            for key in rank_by_dims.keys():
                allowed_values = allowed_values + f'allowed_value: {{ label: "{key}" value: "{key}"}}'

            v + f'''
                parameter: {dynamic_dim_selector} {{
                    type: unquoted

                    {allowed_values}
                }}'''

            v + f'''
                dimension: {dynamic_dim_name} {{
                    type: string
                    hidden: yes
                    sql: {dynamic_dim_sql};;
                }}

            '''
            #create the aggregate ndt
            agg = lookml.View(agg_view)
            agg + f'''
                derived_table: {{
                    explore_source: {e.name} {{
                    column: {dynamic_dim_name} {{field: {v.name}.{dynamic_dim_name}}}
                    column: {rank_by_meas} {{field: {v.name}.{rank_by_meas}}}
                    derived_column: rank {{
                        sql: ROW_NUMBER() OVER (ORDER BY {rank_by_meas} DESC) ;;
                    }}
                    # bind_all_filters: yes
                    bind_filters: {{
                        from_field: {v.name}.{dynamic_dim_selector}
                        to_field: {v.name}.{dynamic_dim_selector}
                    }}
                    }}
                }}
                dimension: {dynamic_dim_name} {{
                    sql: ${{TABLE}}.{dynamic_dim_name} ;;
                    hidden: yes
                }}
                dimension: rank {{
                    type: number
                    hidden: yes
                }}
                filter: tail_threshold {{
                    type: number
                    hidden: yes
                }}
                dimension: stacked_rank {{
                    type: string
                    sql:
                            CASE
                            WHEN ${{rank}} < 10 then '0' || ${{rank}} || ') '|| ${{{dynamic_dim_name}}}
                            ELSE ${{rank}} || ') ' || ${{{dynamic_dim_name}}}
                            END
                    ;;
                }}
                dimension: ranked_by_with_tail {{
                    type: string
                    sql:
                        CASE WHEN {{% condition tail_threshold %}} ${{rank}} {{% endcondition %}} THEN ${{stacked_rank}}
                        ELSE 'x) Other'
                        END
                    ;;
                }}
            '''
            #add our new aggregate view to the view file
            vf + agg
            #join in our aggregate table to the explore
            e + f'''
              join: {agg.name} {{
                type: left_outer
                relationship: many_to_one
                sql_on: ${{{v.name}.{dynamic_dim_name}}}  = ${{{agg.name}.{dynamic_dim_name}}};;
             }}
            '''

            #### SAVE ####
            print(vf)
            vf.write()
            mf.write()
            # p.put(vf)
            # p.put(mf)
            p.deploy()
        apply_top_n(
             lookml.Project(**config['project1'])
            ,'order_items.view.lkml' #ordinarily should be my view file
            ,'order_items'
            ,{'Brand':'${products.brand}','Category':'${products.category}','State':'${users.state}'}
            ,'total_sale_price'
            ,'order_items.model.lkml' 
            ,'order_items'
            )


"""Tokens used by the lexer to tokenize LookML."""


class Token:
    """Base class for LookML tokens, lexed from LookML strings."""

    id: str = "<base token>"
    value: str

    def __init__(self, line_number: int):
        """Initializes a Token.

        Args:
            line_number: The corresponding line in the text where this token begins

        """
        self.line_number: int = line_number

    def __eq__(self, other):
        """Compare one Token to another by their type."""
        return self.__class__ == other.__class__

    def __repr__(self):
        """Returns the token's string representation, truncated to 25 characters.

        If the token has a `value` attribute, include that in the output.

        Examples:
            >>> token = Token(1)
            >>> token.__repr__()
            'Token()'

            >>> token.value = 'A string value'
            >>> token.__repr__()
            'Token(A string value)'

            >>> token.value = 'A very, very, very long string value'
            >>> token.__repr__()
            'Token(A very, very, very long s ... )'

        """
        value = getattr(self, "value", "").strip()
        value = (value[:25].rstrip() + " ... ") if len(value) > 25 else value
        return f"{self.__class__.__name__}({value})"


class ContentToken(Token):
    """Base class for LookML tokens that contain a string of content."""

    def __init__(self, value: str, line_number: int):
        """Initializes a ContentToken with string content.

        Args:
            value: A string value for the token's content
            line_number: The corresponding line in the text where this token begins

        """
        self.value: str = value
        self.line_number: int = line_number

    def __eq__(self, other):
        """Compare one ContentToken to another by their values."""
        return self.id == other.id and self.value == other.value


class StreamStartToken(Token):
    """Represents the start of a stream of characters."""

    id = "<stream start>"


class StreamEndToken(Token):
    """Represents the end of a stream of characters."""

    id = "<stream end>"


class BlockStartToken(Token):
    """Represents the start of a block."""

    id = "{"


class BlockEndToken(Token):
    """Represents the end of a block."""

    id = "}"


class ValueToken(Token):
    """Separates a key from a value."""

    id = ":"


class ExpressionBlockEndToken(Token):
    """Represents the end of an expression block."""

    id = ";;"


class CommaToken(Token):
    """Separates elements in a list."""

    id = ","


class ListStartToken(Token):
    """Represents the start of a list."""

    id = "["


class ListEndToken(Token):
    """Represents the end of a list."""

    id = "]"


class ExpressionBlockToken(ContentToken):
    """Contains the value of an expression block."""

    id = "<expression block>"


class LiteralToken(ContentToken):
    """Contains the value of an unquoted literal."""

    id = "<literal>"


class QuotedLiteralToken(ContentToken):
    """Contains the value of a quoted literal."""

    id = "<quoted literal>"


class RefinementToken(Token):
    """Indicates a LookML refinement."""

    id = "+"



import pathlib
from urllib.parse import urlparse

class url(object):
    def __init__(self,url):
        self.url = urlparse(url)
    def with_port_if_exists(self): 
        return f"{self.url.scheme}://{self.url.hostname}{ ':'+str(self.url.port) if self.url.port else ''}"
    def with_no_port(self): return f"{self.url.scheme}://{self.url.hostname}"

class path(object):
    pass


"""Defines constant sequences of LookML keys."""

from typing import Dict, Tuple, Type

from lookml.lkml import tokens

# These are repeatable keys in LookML that the parser should collapse into a single
# Python dictionary key. For example, LookML can have multiple dimensions, so the parser
# will combine those dimensions into a list of dictionaries with a top-level key,
# `dimensions`.

PLURAL_KEYS: Tuple[str, ...] = (
    "view",
    "measure",
    "dimension",
    "filter",
    "dimension_group",
    "access_filter",
    "bind_filter",
    "map_layer",
    "parameter",
    "set",
    "column",
    "derived_column",
    "include",
    "explore",
    "link",
    "when",
    "allowed_value",
    "named_value_format",
    "join",
    "datagroup",
    "access_grant",
    "sql_step",
    "action",
    "param",
    "form_param",
    "option",
    "user_attribute_param",
    "assert",
    "test",
    "aggregate_table",
    "application",
    "visualization",
    "override_constant",
    "constant",
    "local_dependency",
    "remote_dependency",
    "query",
)

# These are keys in LookML that should be recognized as expression blocks (end with ;;).

EXPR_BLOCK_KEYS: Tuple[str, ...] = (
    "expression_custom_filter",
    "expression",
    "html",
    "sql_trigger_value",
    "sql_table_name",
    "sql_distinct_key",
    "sql_start",
    "sql_always_having",
    "sql_always_where",
    "sql_trigger",
    "sql_foreign_key",
    "sql_where",
    "sql_end",
    "sql_create",
    "sql_latitude",
    "sql_longitude",
    "sql_step",
    "sql_on",
    "sql",
)

# These are keys that the serializer should quote the value of (e.g. `label: "Label"`).
# An example of an unquoted literal would be `hidden: no`.

QUOTED_LITERAL_KEYS: Tuple[str, ...] = (
    "label",
    "view_label",
    "group_label",
    "group_item_label",
    "suggest_persist_for",
    "default_value",
    "direction",
    "value_format",
    "name",
    "url",
    "icon_url",
    "form_url",
    "default",
    "tags",
    "value",
    "description",
    "sortkeys",
    "indexes",
    "partition_keys",
    "connection",
    "include",
    "max_cache_age",
    "allowed_values",
    "timezone",
    "persist_for",
    "cluster_keys",
    "distribution",
    "extents_json_url",
    "feature_key",
    "file",
    "property_key",
    "property_label_key",
    "else",
)

# These are keys for fields in Looker that have a "name" attribute. Since lkml uses the
# key `name` to represent the name of the field (e.g. for `dimension: dimension_name {`,
# the `name` key would hold the value `dimension_name`.)

KEYS_WITH_NAME_FIELDS: Tuple[str, ...] = (
    "user_attribute_param",
    "param",
    "form_param",
    "option",
)

CHARACTER_TO_TOKEN: Dict[str, Type[tokens.Token]] = {
    "\0": tokens.StreamEndToken,
    "{": tokens.BlockStartToken,
    "}": tokens.BlockEndToken,
    "[": tokens.ListStartToken,
    "]": tokens.ListEndToken,
    ",": tokens.CommaToken,
    ":": tokens.ValueToken,
    ";": tokens.ExpressionBlockEndToken,
}




from lookml import core as lookml
from lookml import lkml
import time, subprocess, os
import github
import base64
import requests
import re, shutil
import warnings
import collections
from typing import NewType, Any, Generator, Union, Tuple, Type
# import platform
# import pprint

#P1 simplify code with pathlib
# import pathlib

LOOKML_DASHBOARDS = False
#import marko

def _optional_import_(module: str, name: str = None, package: str = None):
    import importlib
    try:
        module = importlib.import_module(module)
        return module if name is None else getattr(module, name)
    except ImportError as e:
        if package is None:
            package = module
        msg = f"install the '{package}' package to make use of this feature"
        import_error = e

        def _failed_import(*args):
            raise ValueError(msg) from import_error

        return _failed_import


yaml_load = _optional_import_('yaml', 'safe_load', package='pyyaml')
yaml_dump = _optional_import_('yaml', 'dump', package='pyyaml')

class File(object):
    def __init__(self, path: str,):
        self.path = path

        self.lookml_path = self.path

        self.name = self.path.split('/')[-1]

        self._paths_build()

        self._type_determination()

        self.read()

    def _type_determination(self):
        # P1 finish json, md and JS handlers
        self.extension = self.name.split('.')[-1]
        self.named_extension =\
            self.name.split('.')[-2] + '.' + self.name.split('.')[-1]
        if self.extension == 'lkml':
            typ = self.name.split('.')[-2]
            if typ in ('model', 'manifest'):
                self.type = typ
            # views or naked are "partial models"
            else:
                self.type = 'partial_model'
        elif self.extension == 'json':
            self.type = 'json'
        elif self.extension == 'md':
            self.type = 'markdown'
        elif self.extension == 'lookml':
            typ = self.name.split('.')[-2]
            self.type = 'lookml_dashboard'

    def _paths_build(self, creates_folder: bool=True):
        self.python_path = self.path


    def __getattr__(self, item):
        if item in self.__dict__.keys():
            return self.__dict__[item]
        elif item in self.content.__dict__.keys():
            return self.content.__dict__[item]
        else:
            raise Exception(f'{item} not in file type {self.type}')

    def __getitem__(self, item):
        if item in self.content.__dict__.keys():
            return self.content.__dict__[item]

    def __contains__(self, item):
        return item in self.content

    def __add__(self, other):
        self.content + other

    def __sub__(self, other):
        self.content - other

    def read(self):
        if os.path.exists(self.python_path):
            if self.type in ('model', 'partial_model'):
                self.content = lookml.Model(
                    lkml.load(open(self.python_path, encoding="utf-8")), parent=self)
            elif self.type == 'manifest':
                self.content = lookml.Manifest(
                    lkml.load(open(self.python_path, encoding="utf-8")), parent=self)
            elif self.type == 'lookml_dashboard':
                self.content = yaml_load(open(self.python_path,encoding="utf-8"))
                self.content = yaml_load(
                    open(self.python_path, encoding="utf-8"))
        # blank openers for new file
        elif self.type in ('model', 'partial_model'):
            self.content = lookml.Model('', parent=self)
        elif self.type in ('manifest'):
            self.content = lookml.Manifest('', parent=self)
        elif self.type == 'lookml_dashboard':
            self.content = yaml_load(open(self.python_path, encoding="utf-8"))
        else:
            # P1 create other type initialization
            raise Exception('not implemented yet')

    def write(self, overWriteExisting=True):
        """
        Write the file (will write to filesystem or github). It does not write to the remote until 
        .commit() in SSH mode

        Args:
            overWriteExisting (bool, optional): [description]. Defaults to True.
        """
        # print("Writing to: %s" % (self.path) )
        if overWriteExisting:
            with open(self.python_path, 'w') as opened_file:
                try:
                    if self.type == 'lookml_dashboard':
                        warnings.warn('''LookML dashboards are currently read only, 
                        to be fixed in a future release''')
                        # P1 fix writing, currently lookml dashboards should be considered read only
                        # opened_file.write(yaml_dump(self.content))
                    else:
                        opened_file.write(self.__str__())
                except:
                    pass
        else:
            try:
                fh = open(self.python_path, 'r')
                fh.close()
            except FileNotFoundError:
                with open(self.python_path, 'w') as opened_file:
                    opened_file.write(self.__str__())

    def change_name(self):
        """
        Not implemented yet
        """
        #P1: not implemented
        pass

    def move(self):
        """
        Not implemented yet
        """
        #P1: not implemented
        pass

    def delete(self):
        """
        deletes the file from the filesystem
        """
        os.remove(self.python_path)

    def __str__(self): return str(self.content)

class f(File):
    def __init__(
         self 
        ,path: str 
        ,parent=None 
        ,new: bool = True
    ):
        self.parent = parent
        self.path = path
        # lookml_path synonym for code clarity
        self.lookml_path = self.path
        # update the name to the last element of the split
        self.name = self.path.split('/')[-1]

        self._paths_build()
        if new:
            if self.lookml_folder not in self.parent._index.keys():
                self.parent._index[self.lookml_folder] = dict()
            else:
                self.parent._index[self.lookml_folder].update(
                    {
                        self.name: self
                    })
        self._type_determination()

        self.read()

    def _paths_build(self, creates_folder: bool=True):
        # get parent folder
        path_parts = self.path.split('/')
        if len(path_parts) > 1:
            folder = '/'.join(path_parts[:-1])
        else:
            folder = '.'

        # establish python path / pyshical path for write operations
        # vs lookml_path for includes operations
        if not self.path.startswith(self.parent._path):
            folder = self.parent._path + '/' + folder
            self.python_path = self.parent._path + '/' + self.path
        else:
            self.python_path = self.path

        self.lookml_folder = os.path.relpath(folder, self.parent._path)

        # update index
        if self.lookml_folder not in self.parent._index.keys():
            if creates_folder:
                self.parent.check_folder_create(folder)
            self.parent._index[self.lookml_folder] = dict()

    def delete(self):
        # de-ref from index
        del self.parent._index[self.lookml_folder][self.name]
        # perform deletion on filesystem / github
        os.remove(self.python_path)

class f_github(f):
    def __init__(self, githubFile, parent=None, new=False):
        self._github_file = githubFile
        self.parent = parent
        self.path = self._github_file.path
        # lookml_path synonym for code clarity
        self.lookml_path = self.path
        # update the name to the last element of the split
        self.name = self.path.split('/')[-1]

        self._paths_build(creates_folder=False)
        if new:
            if self.lookml_folder not in self.parent._index.keys():
                self.parent._index[self.lookml_folder] = dict()
            else:
                self.parent._index[self.lookml_folder].update(
                    {
                        self.name: self
                    })
        self._type_determination()

        self.read()

    def read(self):
        if isinstance(self._github_file, github.ContentFile.ContentFile):
            if self.type in ('model', 'partial_model'):
                self.content = lookml.Model(lkml.load(base64.b64decode(self._github_file.content).decode('utf-8')), parent=self)
            elif self.type == 'manifest':
                self.content = lookml.Manifest(lkml.load(base64.b64decode(
                    self._github_file.content).decode('utf-8')), parent=self)
            elif self.type == 'lookml_dashboard':
                self.content = yaml_load(base64.b64decode(
                    self._github_file.content).decode('utf-8'))
        # elif isinstance(self._github_file.path, str):
        else:
            if self.type in ('model', 'partial_model'):
                self.content = lookml.Model('', parent=self)
            elif self.type == 'manifest':
                self.content = lookml.Manifest('', parent=self)
            elif self.type == 'lookml_dashboard':
                self.content = yaml_load('')

    def write(self):
        """
        write the file to github

        Raises:
            Exception: if file not in github
        """
        if isinstance(self._github_file.path, str):
            if not self.exists():
                self._github_file = self.parent._git_connection.create_file(
                    self.path,
                    self.parent._commit_message,
                    str(self.content),
                    branch=self.parent._branch
                )
                self._github_file = self._github_file['content']
            else:
                tmp = self.parent._git_connection.get_contents(self.path,ref=self.parent._branch)
                self._github_file = self.parent._git_connection.update_file(
                    self.path,
                    self.parent._commit_message,
                    str(self.content),
                    sha=tmp.sha,
                    branch=self.parent._branch
                )['content']

        elif isinstance(self._github_file, github.ContentFile.ContentFile):
            if self.exists():
                self.parent._git_connection.update_file(
                    self.path,
                    self.parent._commit_message,
                    str(self),
                    sha=self._github_file.sha,
                    branch=self.parent._branch)
        else:
            raise Exception('file was not bound to github or sting path')

    def delete(self):
        """
        Deletes the file from github
        """
        # de-ref from index
        del self.parent._index[self.lookml_folder][self.name]
        # perform deletion on filesystem / github
        tmp = self.parent._git_connection.get_contents(self.path,ref=self.parent._branch)
        self.parent._git_connection.delete_file(
            self.path, self.parent._commit_message, tmp.sha, self.parent._branch
        )

    def exists(self):
        def checkgithub(f0):
            try:
                self.parent._git_connection.get_contents(f0,ref=self.parent._branch)
                return True
            except github.GithubException as e:
                if e._GithubException__status == 404:
                    return False
        return checkgithub(self.path)

class Project(object):
    def __new__(cls, *args, **kwargs):
        # factory sytle routing within class
        if ('repo' in kwargs.keys()) and ('access_token' in kwargs.keys()):
            return super(Project, cls).__new__(ProjectGithub)
        elif ('git_url' in kwargs.keys()):
            return super(Project, cls).__new__(ProjectSSH)
        elif ('path' in kwargs.keys()):
            return super(Project, cls).__new__(cls)

    def __init__(
        self, path: str = '.tmp', 
        git_url: str = '', 
        branch: str = 'master', 
        commitMessage: str = ''
    ):
        self._path = path
        self._index = dict()
        self._build_index()

    def _build_index(self):
        for root, dirs, files in os.walk(self._path, topdown=False):
            for name in files:
                if name.endswith('.lkml') or (name.endswith('.lookml') and LOOKML_DASHBOARDS):
                    folder = os.path.relpath(root, self._path)
                    if folder not in self._index.keys():
                        self._index[folder] = dict()

                    self._index[folder].update(
                        {
                            name: f(
                                os.path.join(
                                    folder,
                                    name
                                ), parent=self, new=False
                            )
                        }
                    )
    @staticmethod
    def _folder_split(f:str):
        Folder = collections.namedtuple('folder',['folder','name','path'])
        if len(f.split('/')) == 1:
            me = Folder(folder='.',name=f,path=f)
        else:
            me = Folder(
                     folder='/'.join(f.split('/')[:-1])
                    ,name=f.split('/')[-1]
                    ,path=f
                    )
        return me

    def __getitem__(self, f):
        f = self._folder_split(f)
        if f.folder in self._index.keys():
            if f.name in self._index[f.folder].keys():
                return self._index[f.folder][f.name]
            else:
                raise Exception(f'{f.name} not found in {f.folder}')
        else:
            raise Exception(f'{f.folder} not found')
    # base

    def __iter__(self):
        self._valueiterator = iter(list(self.traverse(self._index)))
        return self
    # base

    def __next__(self):
        try:
            return next(self._valueiterator)
        except:
            raise StopIteration
    # base

    def traverse(self, tree):
        for k, v in tree.items():
            if isinstance(v, dict):
                for k1, v1 in v.items():
                    if isinstance(v1, dict):
                        self.traverse(v1)
                    else:
                        yield v1
            else:
                yield v
    # base

    def put(self, file: File):
        """
        writes a file to your project. synonym of myFile.write()

        Args:
            file (File): the file object to be written

.. code-block:: python

        myFile = lookml.File('foo.view.lkml')
        proj.put(myFile)

        """
        file.write()

    def delete(self, file: File):
        """
        Delete a file from the project

        Args:
            file (File): pass the file object to delete

.. code-block:: python

        myFile = proj.new_file('foo.view.lkml')
        proj.delete(myFile)
        """
        file.delete()

    def file(self, path: str) -> f:
        """
        Return a file for the lookml project file path

        Args:
            path (str): string to the path of the file i.e 'views/order_items.view.lkml'

        Returns:
            file: the pylookml file type
        """
        return self[path]

    def files(self) -> Generator:
        """
        Iterate over the project files

        Yields:
            Generator: file objects
        """
        for f in self:
            yield f

    def view_files(self):
        """
        Iterate over the project view files

        Yields:
            Generator: file objects
        """
        for f in self:
            if f.named_extension == 'view.lkml':
                yield f

    def model_files(self):
        """
        Iterate over the project model files

        Yields:
            Generator: file objects
        """
        for f in self:
            if f.named_extension == 'model.lkml':
                yield f

    def manifest_files(self):
        """
        Iterate over the project manifest files
        (there should only be one per project)

        Yields:
            Generator: file objects
        """
        for f in self:
            if f.named_extension == 'manifest.lkml':
                yield f

    def check_folder_create(self, dir: str = ''):
        os.makedirs(dir, exist_ok=True)

    def new_file(self, path: str = '') -> f:
        """
        Create a new file of any type: view, model etc

        Args:
            path (str, optional): lookml project path you'd like the file created at

        Returns:
            f: returns the new file object

.. code-block:: python

        myFile = proj.new_file('foo.view.lkml')
        myFile + 'view: foo {}'

        """
        return f(path, parent=self)

    def dir_list(self):
        def traverse(tree, i=0):
            for k, v in tree.items():
                if type(v) == dict:
                    print('   folder: ', k)
                    traverse(v, i=i)
                else:
                    print(' '*6, v.name)
                    if v.type in ('partial_model', 'model'):
                        for vo in v.views:
                            print(' '*9, 'view: ', vo.name)
                        for e in v.explores:
                            print(' '*9, 'explore: ', e.name)
        print('project: ', self._path)
        traverse(self._index)
    # filesystem

    def path_exists(self, path):
        return os.path.exists(path)
    # base
    # filesystem

    def delete_file(self, path:str):
        """
        Delete a file by providing the path (vs delete() which takes a file obj)

        Args:
            fl (str): the path within the project of the file to be deleted

.. code-block:: python

        myFile = proj.new_file('foo.view.lkml')
        proj.delete_file(path='foo.view.lkml')

        """
        self[fl].delete()

#P1: build the ability to delete folders from filesystem & github
class ProjectGithub(Project):
    def __init__(
         self
        ,repo: str = ''
        ,access_token: str = ''
        ,commitMessage: str = ''
        ,branch: str = 'master'
        ,index_whole: bool = False
        ,looker_host: str = ''
        ,looker_project_name: str = ''
        ):
        self._repo = repo
        self._path = self._repo
        self._access_token = access_token
        self._branch = branch
        self._commit_message = "PyLookML Auto Updated: " +\
            time.strftime(
                '%h %d %Y @ %I:%M%p %Z') if not commitMessage else commitMessage
        self.gitsession = github.Github(self._access_token)
        self._git_connection = self.gitsession.get_repo(self._repo)
        self._looker_host = looker_host
        self._looker_project_name = looker_project_name
        self.constructDeployUrl()
        self._index = dict()
        self._index_whole = index_whole
        self._built = False
        if self._index_whole:
            self._build_index()

    def _build_index(self):
        contents = self._git_connection.get_contents('',ref=self._branch)
        while contents:
            file_content = contents.pop(0)
            folder = '/'.join(file_content.path.split('/')[:-1])
            if not folder:
                folder = '.'
            if file_content.type == "dir":
                contents.extend(
                    self._git_connection.get_contents(file_content.path,ref=self._branch))
            else:
                if file_content.name.endswith('.lkml') or \
                        (file_content.name.endswith('.lookml') and LOOKML_DASHBOARDS):
                    if folder not in self._index.keys():
                        self._index[folder] = dict()
                    if file_content.name not in self._index[folder].keys():
                        # self._index[folder][file_content.name] = file_content
                        self._index[folder][file_content.name] = f_github(
                            file_content, parent=self, new=False)
        self._built = True

    def __getitem__(self, f):
        f = self._folder_split(f)
        if f.folder in self._index.keys():
            if f.name in self._index[f.folder].keys():
                return self._index[f.folder][f.name]
            else:
                if not self._built:
                    return f_github(
                        self._git_connection.get_contents(f.path,ref=self._branch), 
                        parent=self, 
                        new=True
                        )
                else:
                    raise Exception(f'{f.name} not found in {f.folder}')
        else:
            if not self._built:
                return f_github(
                    self._git_connection.get_contents(f.path,ref=self._branch), 
                    parent=self, 
                    new=True
                    )
            else:
                raise Exception(f'{f.folder} not found')

    def _exists(self,p:str):
        try:
            self._git_connection.get_contents(p, ref=self._branch)
            return True
        except github.GithubException as e:
            if e._GithubException__status == 404:
                return False

    def new_file(self, path: str = '') -> f:
        path = self._folder_split(path)
        if not self._exists(path.path):
            if path.folder not in self._index.keys():
                self._index[path.folder] = dict()
            if path.name not in self._index[path.folder].keys():
                # self._index[path.folder][path.name] = f_github(path, parent=self, new=True)
                self._index[path.folder][path.name] = f_github(path, parent=self, new=False)
        else:
            raise Exception(f'File {path} already exists,'
                            f' call proj.file("{path}") to modify it'
                            f' or call proj.delete("{path}") to delete it'
                                )
        return self._index[path.folder][path.name]
    def constructDeployUrl(self):
        '''
            Constructs a github deploy URL according to this pattern:
            https://prod.host.com/webhooks/projects/projectname/deploy
        '''
        # P1: the base url should not have had a trailing slash
        if self._looker_project_name and self._looker_host:
            self._deploy_url = \
                self._looker_host + 'webhooks/projects/' + self._looker_project_name + '/deploy'
        else:
            self._deploy_url = ''
    # base
    def put(self, f):
        if type(f) == File:
            new = self.new_file(f.path)
            new.write()
        if isinstance(f,f_github):
            f.write()

    def deploy(self):
        """
        Issues a webhook to have looker sync with your project remote

.. note:: this will only result in changes being displayed if you are deploying to master

.. note:: this will only work if you provided the looker_project_name and looker_host arguments when you initialized the project

        """
        # P3: check to see if project is on master,
        # if not issue warning that changes won't show up
        if self._deploy_url:
            requests.get(self._deploy_url)

class ProjectSSH(Project):
    class gitController:
        def __init__(
            self, outputPath='', projectName='', branch='master', deployMessage='', exePath='', includeGitDir=False
        ):
            # self.platform = platform.system()
            self.platform = 'linux'
            self.preamble = []
            self.trailers = []
            if self.platform == 'Windows':
                self.exe = exePath + ' '
                self.preamble.append(self.exe)
                self.trailers.append(' & exit')
            else:
                self.exe = ''
            self.projectName = projectName

            os.makedirs(outputPath, exist_ok=True)

            self.outputPath = outputPath
            self.absoluteOutputPath = os.path.abspath(self.outputPath)
            self.branch = branch
            self.deployMessage = deployMessage

            if os.path.exists(self.absoluteOutputPath):
                shutil.rmtree(self.absoluteOutputPath)

            os.makedirs(self.absoluteOutputPath, exist_ok=True)

            if self.projectName:
                self.gitDir = ' --git-dir="' + \
                    os.path.abspath(outputPath + '/' +
                                    self.projectName + '/.git') + '" '
            else:
                self.gitDir = ' --git-dir="' + \
                    os.path.abspath(outputPath + '/.git') + '" '

            self.includeGitDir = includeGitDir

            if self.includeGitDir:
                self.preamble.append(self.includeGitDir)

        def call(self, command, gitDir=True):
            if gitDir:
                tmp = ' '.join(self.preamble) + 'git ' + self.gitDir + \
                    ' ' + command + ' '.join(self.trailers)
            else:
                tmp = ' '.join(self.preamble) + 'git ' + \
                    command + ' '.join(self.trailers)
            print(tmp)
            proc = subprocess.Popen(
                tmp, shell=True, env=os.environ, cwd=self.absoluteOutputPath
            )
            try:
                outs, errs = proc.communicate(timeout=15)
            except subprocess.TimeoutExpired:
                proc.kill()
                outs, errs = proc.communicate()
            return self

        def pull(self):
            return self.call(' pull origin ' + self.branch + ' ', gitDir=False)

        def clone(self, repoLocation):
            self.call(' clone ' + repoLocation + ' ' +
                      self.absoluteOutputPath, gitDir=False)
            return self.pull()

        def add(self, path='.'):
            return self.call(' add ' + path, gitDir=False)

        def commit(self, message=''):
            if message:
                return self.call(' commit -m "' + message + ' ' + self.deployMessage + '"', gitDir=False)
            else:
                return self.call(' commit -m "' + self.deployMessage + '"', gitDir=False)

        def pushRemote(self):
            return self.call(' push origin ' + self.branch + ' ', gitDir=False)

    def __init__(
        self,
        git_url: str,
        looker_project_name: str, 
        path: str = '.tmp', 
        looker_host: str = None,  
        branch: str = 'master', 
        commitMessage: str = ''
        ):
        self._path = path
        self._index = dict()
        self._looker_host = looker_host
        self._looker_project_name = looker_project_name
        self.constructDeployUrl()
        self._git_url = git_url
        self._branch = branch
        self._commit_message = "PyLookML Auto Updated: " +\
            time.strftime(
                '%h %d %Y @ %I:%M%p %Z') if not commitMessage else commitMessage
        if self._git_url:
            self._path = self._path + '/' + self._looker_project_name
            self._git = self.gitController(
                outputPath=self._path, 
                projectName=self._looker_project_name, 
                branch=self._branch, 
                deployMessage=self._commit_message
            )
            self._git.clone(self._git_url)
        self._build_index()
    # shell
    def commit(self):
        """
        git commit and git push changes to remote in SSH mode
        """
        self._git.commit()
        self._git.pushRemote()

    def constructDeployUrl(self):
        '''
            Constructs a github deploy URL according to this pattern:
            https://prod.host.com/webhooks/projects/projectname/deploy
        '''
        # P1: the base url should not have had a trailing slash
        if self._looker_project_name and self._looker_host:
            self._deploy_url = \
                self._looker_host + 'webhooks/projects/' + self._looker_project_name + '/deploy'
        else:
            self._deploy_url = ''
    # base

    def deploy(self):
        """
        Issues a webhook to have looker sync with your project remote

.. note:: this will only result in changes being displayed if you are deploying to master

.. note:: this will only work if you provided the looker_project_name and looker_host arguments when you initialized the project

        """
        if self._deploy_url:
            requests.get(self._deploy_url)



import argparse
import io
import json
import logging
import sys
from typing import IO, Optional, Sequence, Union

from lookml.lkml.lexer import Lexer
from lookml.lkml.parser import Parser
from lookml.lkml.serializer import Serializer

def load(stream: Union[str, IO]) -> dict:
    """Parse LookML into a Python dictionary.

    Args:
        stream: File object or string containing LookML to be parsed

    Raises:
        TypeError: If stream is neither a string or a file object

    """

    if isinstance(stream, io.TextIOWrapper):
        text = stream.read()
    elif isinstance(stream, str):
        text = stream
    else:
        raise TypeError("Input stream must be a string or file object.")
    lexer = Lexer(text)
    tokens = lexer.scan()
    parser = Parser(tokens)
    result = parser.parse()
    return result


def dump(obj: dict, file_object: IO = None) -> Optional[str]:
    """Serialize a Python dictionary into LookML.

    Args:
        obj: The Python dictionary to be serialized to LookML
        file_object: An optional file object to save the LookML string to

    Returns:
        A LookML string if no file_object is passed

    """
    serializer = Serializer()
    result = serializer.serialize(obj)
    if file_object:
        file_object.write(result)
        return None
    else:
        return result


def parse_args(args: Sequence) -> argparse.Namespace:
    """Parse command-line arguments."""
    parser = argparse.ArgumentParser(
        description=(
            "A blazing fast LookML parser, implemented in pure Python. "
            "When invoked from the command line, "
            "returns the parsed output as a JSON string."
        )
    )
    parser.add_argument(
        "file", type=argparse.FileType("r"), help="path to the LookML file to parse"
    )
    parser.add_argument(
        "-d",
        "--debug",
        action="store_const",
        dest="log_level",
        const=logging.DEBUG,
        default=logging.WARN,
        help="increase logging verbosity",
    )

    return parser.parse_args(args)


def cli():
    """Command-line entry point for lkml."""
    logger = logging.getLogger()
    logger.setLevel(logging.WARN)

    handler = logging.StreamHandler()
    handler.setLevel(logging.DEBUG)

    formatter = logging.Formatter("%(name)s %(levelname)s: %(message)s")

    handler.setFormatter(formatter)
    logger.addHandler(handler)

    args = parse_args(sys.argv[1:])

    logging.getLogger().setLevel(args.log_level)

    lookml = load(args.file)
    args.file.close()

    json_string = json.dumps(lookml, indent=2)
    print(json_string)


if __name__ == "__main__":
    cli()




import src.lookml.lookml as lookml
import src.lookml.lkml as lkml

import unittest, copy
from pprint import pprint
import configparser, json
from looker_sdk import models, methods, init31
config = configparser.ConfigParser()
config.read('tests/settings.ini')

class testScenariosDeployExtension(unittest.TestCase):
    '''
        Objective: test end to end common use case scenarios 
    '''
    def setUp(self):
        pass
    def test_deploy_extension(self):
        #deploy a parametarized extension
        pass

